# Copyright 2024 CrackNuts. All rights reserved.

from cracknuts.cracker.cracker_s1 import CrackerS1
from cracknuts.cracker.cracker_g1 import CrackerG1


def cracker_s1(
        address: tuple | str,
        bin_server_path: str | None = None,
        bin_bitstream_path: str | None = None
):
    """
    :param address: Cracker 设备地址，可以是：
                        - IP 和端口的元组 `(ip, port)`
                        - 字符串 `"[cnp://]<ip>[:port]"`
                    例如：
                        - "192.168.0.10"
                        - "cnp://192.168.0.10:9761"
                        - (192.168.0.10, 9761)
    :type address: str | tuple | None
    :param bin_server_path: bin_server（固件）文件的路径，用于更新；通常情况下用户不需要指定。
    :type bin_server_path: str | None
    :param bin_bitstream_path: bin_bitstream（固件）文件的路径，用于更新；通常情况下用户不需要指定。
    :type bin_bitstream_path: str | None
    """
    return CrackerS1(address, bin_server_path, bin_bitstream_path)


def cracker_g1(
        address: tuple | str,
        bin_server_path: str | None = None,
        bin_bitstream_path: str | None = None
):
    """
    :param address: Cracker 设备地址，可以是：
                        - IP 和端口的元组 `(ip, port)`
                        - 字符串 `"[cnp://]<ip>[:port]"`
                    例如：
                        - "192.168.0.10"
                        - "cnp://192.168.0.10:9761"
                        - (192.168.0.10, 9761)
    :type address: str | tuple | None
    :param bin_server_path: bin_server（固件）文件的路径，用于更新；通常情况下用户不需要指定。
    :type bin_server_path: str | None
    :param bin_bitstream_path: bin_bitstream（固件）文件的路径，用于更新；通常情况下用户不需要指定。
    :type bin_bitstream_path: str | None
    """
    return CrackerG1(address, bin_server_path, bin_bitstream_path)


__all__ = ["cracker_s1", "cracker_g1"]



# Copyright 2024 CrackNuts. All rights reserved.

import abc
import enum
import sqlite3
import typing
from abc import ABC

from cracknuts.acquisition.acquisition import Acquisition
from cracknuts.acquisition.acquisition import AcquisitionBuilder
from cracknuts.cracker.cracker_basic import CrackerBasic
from cracknuts.cracker.cracker_g1 import CrackerG1
from cracknuts.glitch.param_generator import (
    AbstractGlitchParamGenerator,
    VCCGlitchParam,
    VCCGlitchParamGenerator,
    GlitchGenerateParam,
    GNDGlitchParamGenerator,
)


class GlitchTestResult(abc.ABC):
    def __init__(self, path: str, create=True):
        self._path = path
        self._conn = sqlite3.connect(path)
        self._cursor = self._conn.cursor()
        if create:
            self._init_table()

    def _init_table(self): ...

    def add(self, index, param, data: "GlitchDoData"): ...

    def close(self):
        self._cursor.close()
        self._conn.close()

    def open(self):
        self._conn = sqlite3.connect(self._path)
        self._cursor = self._conn.cursor()


class VCCGlitchTestResult(GlitchTestResult):
    def _init_table(self):
        self._cursor.execute("""
        CREATE TABLE IF NOT EXISTS glitch_result (
            `no` INTEGER PRIMARY KEY,
            normal INTEGER,
            wait INTEGER,
            glitch INTEGER,
            `count` INTEGER,
            repeat INTEGER,
            `interval` INTEGER,
            plaintext BLOB,
            ciphertext BLOB,
            `key` BLOB,
            extended BLOB,
            status INTEGER
        )
        """)
        self._conn.commit()
        self._conn.execute("PRAGMA journal_mode=WAL;")

    def add(self, index, param: VCCGlitchParam, data: "GlitchDoData"):
        if data.glitch_status is None:
            data.glitch_status = GlitchDoStatus.ERROR
        data.key = data.key.hex() if data.key is not None else ""
        data.ciphertext = data.ciphertext.hex() if data.ciphertext is not None else ""
        data.plaintext = data.plaintext.hex() if data.plaintext is not None else ""
        data.extended = data.extended.hex() if data.extended is not None else ""
        self._cursor.execute(
            """
                INSERT INTO glitch_result (
                    no, normal, wait, glitch, count, repeat, interval, plaintext, ciphertext, key, extended, status
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
            (
                index,
                param.normal,
                param.wait,
                param.glitch,
                param.count,
                param.repeat,
                param.interval,
                data.plaintext,
                data.ciphertext,
                data.key,
                data.extended,
                data.glitch_status.value,
            ),
        )
        self._conn.commit()


class GlitchAcquisition(Acquisition, ABC):
    def __init__(
        self,
        cracker: CrackerG1,
        trace_count: int = 1000,
        shadow_trace_count: int = 1000,
        sample_length: int = -1,
        sample_offset: int = 0,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        trigger_judge_wait_time: float = 0.001,
        trigger_judge_timeout: float = 1.0,
        do_error_handler_strategy: int = Acquisition.DO_ERROR_HANDLER_STRATEGY_EXIT,
        do_error_max_count: int = -1,
        file_format: str = "zarr",
        file_path: str = "auto",
        trace_fetch_interval: float = 0,
    ):
        super().__init__(
            cracker,
            trace_count,
            sample_length,
            sample_offset,
            data_plaintext_length,
            data_ciphertext_length,
            data_key_length,
            data_extended_length,
            trigger_judge_wait_time,
            trigger_judge_timeout,
            do_error_handler_strategy,
            do_error_max_count,
            file_format,
            file_path,
            trace_fetch_interval,
        )
        self._shadow_trace_count = shadow_trace_count
        self.cracker: CrackerG1 = cracker
        self._glitch_result = None
        self._glitch_param_generator: AbstractGlitchParamGenerator | None = None
        self._current_glitch_param = None
        self._is_in_glitch_mode = False
        self._glitch_test_params = None

        self._on_glitch_result_added_listener = []

    @property
    def shadow_trace_count(self):
        return self._shadow_trace_count

    @shadow_trace_count.setter
    def shadow_trace_count(self, value: int):
        self._shadow_trace_count = value
        for listener in self._on_config_changed_listener:
            listener("shadow_trace_count", value)

    def pre_init(self):
        if not self._is_in_glitch_mode:
            return
        if isinstance(self._glitch_param_generator, VCCGlitchParamGenerator):
            self._glitch_result = VCCGlitchTestResult(f"dataset/{self.current_timestamp}.db", create=True)
        else:
            # todo other glitch test result.
            ...

    def do(self, count: int) -> "GlitchDoData": ...

    def pre_do(self):
        if self._is_in_glitch_mode:
            if isinstance(self._glitch_param_generator, VCCGlitchParamGenerator):
                glitch_param = self._glitch_param_generator.next()
                cracker_g1: CrackerG1 = typing.cast(CrackerG1, self.cracker)
                cracker_g1.glitch_vcc_config(
                    wait=glitch_param.wait,
                    level=glitch_param.glitch,
                    length=glitch_param.count,
                    delay=glitch_param.interval,
                    repeat=glitch_param.repeat,
                )
                self._current_glitch_param = glitch_param
            # OTHER GLITCH PARAM GENERATOR
        self.cracker.osc_single()

    def _post_do(self, index, data):
        if self._is_in_glitch_mode and self._glitch_result is not None:
            data = GlitchDoData(**data)
            self._glitch_result.add(index, self._current_glitch_param, data)
            for listener in self._on_glitch_result_added_listener:
                listener(index, self._current_glitch_param, data)

    def _pre_finish(self):
        if self._is_in_glitch_mode and self._glitch_result is not None:
            self._glitch_result.close()

    def set_glitch_params(self, param_generator: AbstractGlitchParamGenerator):
        self._glitch_param_generator = param_generator
        # self.trace_count = self._glitch_param_generator.total()

    def test(
        self,
        # count=-1,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        trace_fetch_interval: float = 2.0,
    ):
        self._is_in_glitch_mode = False
        super().test(
            # count=count,
            sample_length=sample_length,
            sample_offset=sample_offset,
            trigger_judge_wait_time=trigger_judge_wait_time,
            trigger_judge_timeout=trigger_judge_timeout,
            do_error_max_count=do_error_max_count,
            do_error_handler_strategy=do_error_handler_strategy,
            trace_fetch_interval=trace_fetch_interval,
        )

    # def glitch_test(
    #     self,
    #     sample_length: int | None = None,
    #     sample_offset: int | None = None,
    #     trigger_judge_wait_time: float | None = None,
    #     trigger_judge_timeout: float | None = None,
    #     do_error_max_count: int | None = None,
    #     do_error_handler_strategy: int | None = None,
    #     trace_fetch_interval: float = 0.1,
    # ):
    #     glitch_params = self._build_glitch_param_generator(self._glitch_test_params)
    #     self._logger.warning(f"glitch_params: {glitch_params}")
    #     self.set_glitch_params(glitch_params)
    #     self._logger.warning(f"count is {self.trace_count}")
    #
    #     test_mode = self._glitch_param_generator.total() == 1
    #     count = -1 if test_mode else self._glitch_param_generator.total()
    #
    #     self._is_in_glitch_mode = True
    #     self._run(
    #         test=test_mode,
    #         persistent=False,
    #         count=count,
    #         sample_length=sample_length,
    #         sample_offset=sample_offset,
    #         trigger_judge_wait_time=trigger_judge_wait_time,
    #         trigger_judge_timeout=trigger_judge_timeout,
    #         do_error_max_count=do_error_max_count,
    #         do_error_handler_strategy=do_error_handler_strategy,
    #         trace_fetch_interval=trace_fetch_interval,
    #     )

    def run(
        self,
        count: int = None,
        sample_length: int = None,
        sample_offset: int = 0,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "zarr",
        file_path: str | None = "auto",
    ):
        self._is_in_glitch_mode = False
        super().run(
            count=count,
            sample_length=sample_length,
            sample_offset=sample_offset,
            data_plaintext_length=data_plaintext_length,
            data_ciphertext_length=data_ciphertext_length,
            data_key_length=data_key_length,
            data_extended_length=data_extended_length,
            trigger_judge_wait_time=trigger_judge_wait_time,
            trigger_judge_timeout=trigger_judge_timeout,
            do_error_max_count=do_error_max_count,
            do_error_handler_strategy=do_error_handler_strategy,
            file_format=file_format,
            file_path=file_path,
        )

    def glitch_run(
        self,
        count: int | None = None,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "zarr",
        file_path: str | None = "auto",
        trace_fetch_interval: float = 0.1,
    ):
        glitch_params = self._build_glitch_param_generator(self._glitch_test_params)
        self.set_glitch_params(glitch_params)

        if count is None:
            count = self.shadow_trace_count
        if self._glitch_test_params is not None:
            count = count * self._glitch_param_generator.total()
        self._is_in_glitch_mode = True
        self._run(
            test=False,
            persistent=True,
            count=count,
            sample_length=sample_length,
            sample_offset=sample_offset,
            data_plaintext_length=data_plaintext_length,
            data_ciphertext_length=data_ciphertext_length,
            data_key_length=data_key_length,
            data_extended_length=data_extended_length,
            trigger_judge_wait_time=trigger_judge_wait_time,
            trigger_judge_timeout=trigger_judge_timeout,
            do_error_max_count=do_error_max_count,
            do_error_handler_strategy=do_error_handler_strategy,
            file_format=file_format,
            file_path=file_path,
            trace_fetch_interval=trace_fetch_interval,
            is_glitch=True,
        )

    @staticmethod
    def _build_glitch_param_generator(glitch_params) -> None | VCCGlitchParamGenerator | GNDGlitchParamGenerator:
        if glitch_params is None:
            return None
        if glitch_params["type"] == "vcc":
            normal, wait, glitch, count, repeat, interval = None, None, None, None, None, None
            for param in glitch_params["data"]:
                if param["prop"] == "normal":
                    g_param = param["param"]
                    normal = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "wait":
                    g_param = param["param"]
                    wait = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "glitch":
                    g_param = param["param"]
                    glitch = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "count":
                    g_param = param["param"]
                    count = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "repeat":
                    g_param = param["param"]
                    repeat = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "interval":
                    g_param = param["param"]
                    interval = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
            return VCCGlitchParamGenerator(normal, wait, glitch, count, repeat, interval)
        elif glitch_params["type"] == "gnd":
            normal, wait, glitch, count, repeat, interval = None, None, None, None, None, None
            for param in glitch_params["params"]:
                if param["prop"] == "normal":
                    g_param = param["param"]
                    normal = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "wait":
                    g_param = param["param"]
                    wait = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "glitch":
                    g_param = param["param"]
                    glitch = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "count":
                    g_param = param["param"]
                    count = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "repeat":
                    g_param = param["param"]
                    repeat = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
                elif param["prop"] == "interval":
                    g_param = param["param"]
                    interval = GlitchGenerateParam(
                        start=g_param["start"],
                        end=g_param["end"],
                        count=g_param["count"],
                        step=g_param["step"],
                        mode=GlitchGenerateParam.Mode(g_param["mode"]),
                    )
            return GNDGlitchParamGenerator(normal, wait, glitch, count, repeat, interval)
        elif glitch_params["type"] == "clock":
            ...
        else:
            raise ValueError("Unknown glitch param type")

    def glitch_test_sync(
        self,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "zarr",
        file_path: str | None = "auto",
        trace_fetch_interval: float = 0.1,
    ):
        self._do_run(
            test=False,
            persistent=False,
            count=self.trace_count,
            sample_length=sample_length,
            sample_offset=sample_offset,
            data_plaintext_length=data_plaintext_length,
            data_ciphertext_length=data_ciphertext_length,
            data_key_length=data_key_length,
            data_extended_length=data_extended_length,
            trigger_judge_wait_time=trigger_judge_wait_time,
            trigger_judge_timeout=trigger_judge_timeout,
            do_error_max_count=do_error_max_count,
            do_error_handler_strategy=do_error_handler_strategy,
            file_format=file_format,
            file_path=file_path,
            trace_fetch_interval=trace_fetch_interval,
        )

    def set_glitch_test_params(self, param):
        self._glitch_test_params = param

    def get_glitch_test_params(self):
        return self._glitch_test_params

    def on_glitch_result_added(self, callback):
        self._on_glitch_result_added_listener.append(callback)

    @staticmethod
    def simple(
            cracker: CrackerG1,
            init_func: typing.Callable[[CrackerG1], None] = lambda cracker: None,
            do_func: typing.Callable[[CrackerG1, int], dict[str, bytes]] = lambda cracker, count: {},
            finish_func: typing.Callable[[CrackerG1], None] = lambda cracker: None, **kwargs) -> 'Acquisition':
        """
        创建一个简单的Acquisition子类实例。

        :param cracker: Cracker实例
        :type cracker: CrackerG1
        :param init_func:
            初始化函数，该函数接受一个CrackerG1实例作为参数，用于在采集开始前进行初始化操作。
            参数说明：
                - cracker: (CrackerG1) Cracker实例
                - 返回值：无
            示例：
                cracker.nut_voltage_enable()
                cracker.nut_voltage(3.3)
                cracker.nut_clock_enable()
                cracker.nut_clock_freq('8M')
                cracker.uart_io_enable()
                status, ret = cracker.uart_transmit_receive(cmd_set_aes_enc_key + aes_key, timeout=1000, rx_count=6)
        :type init_func: typing.Callable[[CrackerG1], None]
        :param do_func:
            执行函数，具体的采集逻辑实现，该函数接受一个CrackerG1实例和当前采集计数作为参数，返回一个包含采集数据的字典。
            参数说明：
                - cracker: (CrackerG1) Cracker实例
                - count: (int) 当前采集计数，从0开始
                - 返回值：dict[str, bytes]，包含采集数据的字典，格式如下：
                  {
                      "plaintext": 明文数据的字节串,
                      "ciphertext": 密文数据的字节串,
                      "key": 密钥数据的字节串（可选）,
                      "extended": 扩展数据的字节串（可选）,
                  }
            示例：
                def do_func(cracker: CrackerG1, count: int) -> dict[str, bytes]:
                    plaintext = random.randbytes(aes_data_len)
                    status, ciphertext = cracker.uart_transmit_receive(cmd_aes_enc + plaintext_data, rx_count= 12)
                    return {
                        "plaintext": plaintext,
                        "ciphertext": ciphertext,
                        "key": aes_key,
                    }
        :type do_func: typing.Callable[[CrackerG1, int], dict[str, bytes]]
        :param finish_func:
            结束后处理函数，该函数接受一个CrackerG1实例作为参数，用于在采集结束后进行清理操作。
            参数说明：
                - cracker: (CrackerG1) Cracker实例
                - 返回值：无
            示例：
                def finish_func(cracker: CrackerG1) -> None:
                    cracker.nut_voltage_disable()
                    cracker.nut_clock_disable()
                    cracker.uart_io_disable()
        :type finish_func: typing.Callable[[CrackerG1], None]
        :param kwargs: 其他Acquisition的关键字参数
        :type kwargs: dict
        :return: Acquisition实例
        """
        class AnonymousAcquisition(GlitchAcquisition):
            def __init__(self, **_kwargs):
                _kwargs['cracker'] = cracker
                super().__init__(**_kwargs)

            def init(self):
                init_func(cracker)

            def do(self, count: int):
                return do_func(cracker, count)

            def finish(self):
                finish_func(cracker)

        return AnonymousAcquisition(**kwargs)


class GlitchDoStatus(enum.Enum):
    GLITCHED = 0
    NOT_GLITCHED = 1
    NO_RETURN = 2
    ERROR = 3


class GlitchDoData:
    def __init__(
        self,
        plaintext: bytes | None = None,
        ciphertext: bytes | None = None,
        key: bytes | None = None,
        extended: bytes | None = None,
        glitch_status: GlitchDoStatus | int = GlitchDoStatus.NOT_GLITCHED,
    ):
        self.plaintext = plaintext
        self.ciphertext = ciphertext
        self.key = key
        self.extended = extended
        if isinstance(glitch_status, int):
            glitch_status = GlitchDoStatus(glitch_status)
        self.glitch_status = glitch_status


class GlitchAcquisitionBuilder(AcquisitionBuilder):
    def build(self, **kwargs):
        """
        Build the acquisition object.

        :param kwargs: other keyword arguments of Acquisition.
        :type kwargs: dict
        :return Acquisition
        """
        if self._cracker is None:
            raise ValueError("No cracker")

        builder_self = self

        class AnonymousAcquisition(GlitchAcquisition):
            def __init__(self, *ana_args, **ana_kwargs):
                super().__init__(*ana_args, **ana_kwargs)

            def init(self):
                builder_self._init_function()

            def do(self, count: int):
                return builder_self._do_function(count)

            def finish(self):
                builder_self._finish_function()

        return AnonymousAcquisition(self._cracker, **kwargs)



# Copyright 2024 CrackNuts. All rights reserved.

import json
import os
import pathlib
import sys
import typing
from enum import Enum

from traitlets import traitlets

from cracknuts import logger
from cracknuts.acquisition.acquisition import Acquisition
from cracknuts.acquisition.glitch_acquisition import GlitchAcquisition, GlitchDoData, GlitchAcquisitionBuilder
from cracknuts.cracker.cracker_g1 import CrackerG1
from cracknuts.jupyter.acquisition_panel import GlitchAcquisitionPanelWidget
from cracknuts.jupyter.cracker_g1_panel import CrackerG1Panel
from cracknuts.jupyter.glitch_test_panel import GlitchTestPanel
from cracknuts.jupyter.panel import MsgHandlerPanelWidget
from cracknuts.jupyter.scope_panel import ScopePanelWidget
from cracknuts.jupyter.ui_sync import observe_interceptor
from cracknuts.utils import user_config


class WorkbenchG1Panel(
    CrackerG1Panel,
    GlitchAcquisitionPanelWidget,
    ScopePanelWidget,
    MsgHandlerPanelWidget,
):
    _esm = pathlib.Path(__file__).parent / "static" / "CrackerG1Widget.js"
    _css = ""

    language = traitlets.Unicode("en").tag(sync=True)

    glitch_test_params = traitlets.Dict({}).tag(sync=True)

    def __init__(self, cracker: CrackerG1, acquisition: GlitchAcquisition):
        super().__init__(**{"cracker": cracker, "acquisition": acquisition})
        self._logger = logger.get_logger(self)

        self.language = user_config.get_option("language", fallback="en")
        self.glitch_test_panel = GlitchTestPanel()

        self.cracker: CrackerG1 = cracker
        self.acquisition: GlitchAcquisition = acquisition

        self.uri = self.cracker.get_uri()
        self.language = user_config.get_option("language", fallback="en")

        if self.cracker.get_connection_status():
            self._has_connected_before = True
            workspace_config = self._get_workspace_config()
            if workspace_config:
                config_file_cracker_config = workspace_config.get("cracker")
                # uri = workspace_config.get("connection")
                uri = cracker.get_uri()
                cracker_config = cracker.get_current_config()
                if cracker_config is not None:
                    if config_file_cracker_config:
                        for k, v in config_file_cracker_config.items():
                            # Skip comparison for ignored configuration items.
                            if k in ("nut_timeout",):
                                continue
                            if hasattr(cracker_config, k):
                                cv = getattr(cracker_config, k)
                                if isinstance(cv, Enum):
                                    cv = cv.value
                                if v != cv:
                                    self.panel_config_different_from_cracker_config = True
                                    self._logger.warning(
                                        f"The configuration item {k} differs between the configuration file "
                                        f"({v}) and the cracker ({cv})."
                                    )
                                    break
                            else:
                                self._logger.error(
                                    f"Config has no attribute named {k}, "
                                    f"which comes from the JSON key in the config file."
                                )
                        if not self.panel_config_different_from_cracker_config:
                            self.listen_cracker_config()
                        self.update_cracker_panel_config(config_file_cracker_config, uri)
                    else:
                        self._logger.error(
                            "Configuration file format error: The cracker configuration segment is missing. "
                            "The configuration from the cracker or the default configuration will be used."
                        )
                        self.read_config_from_cracker()
                        self.listen_cracker_config()
                else:
                    self.update_cracker_panel_config(config_file_cracker_config, uri)

                acquisition_config = workspace_config.get("acquisition")
                if acquisition_config:
                    self.acquisition.load_config_from_json(workspace_config.get("acquisition"))
                else:
                    self._logger.error(
                        "Configuration file format error: Acquisition configuration segment is missing. "
                        "The configuration from the acquisition object or "
                        "the default configuration will be used."
                    )
                self.sync_config_from_acquisition()
                self.listen_acquisition_config()
            else:
                self.read_config_from_cracker()
                self.listen_cracker_config()
                self.sync_config_from_acquisition()
                self.listen_acquisition_config()

        self.reg_msg_handler("dumpConfigButton", "onClick", self.dump_config_button_click)
        self.reg_msg_handler("loadConfigButton", "onClick", self.load_config_button_click)
        self.reg_msg_handler("saveConfigButton", "onClick", self.save_config_button_click)
        self.reg_msg_handler("writeConfigButton", "onClick", self.write_config_button_click)
        self.reg_msg_handler("readConfigButton", "onClick", self.read_config_button_click)

    def before_test(self):
        self.write_config_to_cracker()

    def before_run(self):
        self.write_config_to_cracker()

    def dump_config_button_click(self, args: dict[str, typing.Any]):
        self.send({"dumpConfigCompleted": self._dump_config()})

    def load_config_button_click(self, args: dict[str, typing.Any]):
        connection_info = args.get("connection")
        config_file_cracker_config = args.get("cracker")
        acquisition_info = args.get("acquisition")
        cracker_config = self.cracker.get_current_config()
        for k, v in config_file_cracker_config.items():
            if hasattr(cracker_config, k):
                cv = getattr(cracker_config, k)
                if isinstance(cv, Enum):
                    cv = cv.value
                if v != cv:
                    self.panel_config_different_from_cracker_config = True
                    self._logger.error(
                        f"The configuration item {k} differs between the configuration file "
                        f"({v}) and the cracker ({cv})."
                    )
                    break
            else:
                self._logger.error(
                    f"Config has no attribute named {k}, " f"which comes from the JSON key in the config file."
                )

        if self.panel_config_different_from_cracker_config:
            self.stop_listen_cracker_config()
            self.update_cracker_panel_config(config_file_cracker_config, connection_info)

        self.acquisition.load_config_from_str(acquisition_info)
        self.read_config_from_cracker()
        self.send({"loadConfigCompleted": True})

    def write_config_button_click(self, args: dict[str, typing.Any]):
        self.write_config_to_cracker()

    def read_config_button_click(self, args: dict[str, typing.Any]):
        self.read_config_from_cracker()

    def save_config_button_click(self, args: dict[str, typing.Any]):
        current_config_path = self._get_workspace_config_path()
        if current_config_path is None:
            return
        with open(current_config_path, "w") as f:
            self._logger.error(self._dump_config())
            f.write(self._dump_config())

    def _dump_config(self):
        def enum_converter(obj):
            if isinstance(obj, Enum):
                return obj.value
            raise TypeError(f"Type {type(obj)} not serializable")

        return json.dumps(
            {
                "connection": self.uri,
                "cracker": self.get_cracker_panel_config().__dict__,
                "acquisition": self.get_acquisition_panel_config().__dict__,
            },
            indent=4,
            default=enum_converter,
        )

    def _load_current_path_config(self):
        """
        Load the workspace configuration if it exists.
        """
        config = self._get_workspace_config()

        if config is not None:
            connection_info = config.get("connection")
            config_info = config.get("cracker")
            acquisition_info = config.get("acquisition")

            self.cracker.load_config_from_str(config_info)
            self.cracker.set_uri(connection_info)
            self.acquisition.load_config_from_str(acquisition_info)
        else:
            self._logger.debug("Current config path is None or is no exist, current config will be ignored.")

    def _get_workspace_config(self):
        current_config_path = self._get_workspace_config_path()
        if current_config_path is None:
            return
        if current_config_path is not None and os.path.exists(current_config_path):
            with open(current_config_path) as f:
                try:
                    config = json.load(f)
                    return config
                except json.JSONDecodeError as e:
                    self._logger.error(f"Load workspace config file failed: {e.args}")
                    return None

    def _get_workspace_config_path(self):
        global_vars = sys.modules["__main__"].__dict__

        ipynb_path = None

        if "__vsc_ipynb_file__" in global_vars:
            ipynb_path = global_vars["__vsc_ipynb_file__"]
        elif "__session__" in global_vars:
            ipynb_path = global_vars["__session__"]

        if ipynb_path is not None:
            config_path = os.path.join(os.path.dirname(ipynb_path), ".config")
            if not os.path.exists(config_path):
                os.makedirs(config_path)
            elif os.path.isfile(config_path):
                self._logger.error(f"The config directory ({config_path}) already exists, and is not a directory.")
                return None
            return os.path.join(os.path.dirname(ipynb_path), ".config", os.path.basename(ipynb_path)[:-6] + ".json")

    @traitlets.observe("language")
    def on_language_change(self, change):
        user_config.set_option("language", change["new"])

    @staticmethod
    def builder():
        return CrackerG1PanelBuilder()

    @traitlets.observe("glitch_test_params")
    @observe_interceptor
    def glitch_test_params_changed(self, change):
        if isinstance(self.cracker, CrackerG1):
            self.acquisition.set_glitch_test_params(change.get("new"))
            # self.acquisition.glitch_run()


class CrackerG1PanelBuilder:
    def __init__(self):
        self._cracker: CrackerG1 | None = None
        self._acquisition_builder: GlitchAcquisitionBuilder | None = None
        self._acquisition_kwargs: dict = {}
        self._acquisition: Acquisition | None = None

    def with_cracker(self, cracker: CrackerG1) -> "CrackerG1PanelBuilder":
        self._cracker = cracker
        if self._acquisition_builder:
            self._acquisition_builder.cracker(cracker)
        return self

    def with_acquisition(
        self,
        init: typing.Callable[[], None],
        do: typing.Callable[[int], GlitchDoData],
        finish: typing.Callable[[], None],
        **kwargs,
    ) -> "CrackerG1PanelBuilder":
        """
        配置Acquisition(采集流程)参数

        :param init: 采集开始前的初始化函数，一般进行一些一次性配置，如电压配置、采样长度等
        :type init: typing.Callable[[], None]
        :param do: 采集函数，参数为采集次数，返回值为GlitchDoData对象
        :type do: typing.Callable[[int], GlitchDoData]
        :param finish: 采集结束后的清理函数，一般进行一些资源释放
        :type finish: typing.Callable[[], None]
        :param kwargs: 其他Acquisition参数
        :return: CrackerG1PanelBuilder对象
        :rtype: CrackerG1PanelBuilder
        """
        self._acquisition_builder = GlitchAcquisitionBuilder()
        self._acquisition_builder.init(init).do(do).finish(finish)
        if self._cracker:
            self._acquisition_builder.cracker(self._cracker)
        self._acquisition_kwargs = kwargs
        return self

    def build(self):
        self._acquisition = self._acquisition_builder.build(**self._acquisition_kwargs)
        return WorkbenchG1Panel(cracker=self._cracker, acquisition=self._acquisition)



# Copyright 2024 CrackNuts. All rights reserved.

import numpy as np
import plotly.graph_objects as go
from numpy.typing import NDArray
from plotly_resampler import FigureWidgetResampler

from cracknuts.trace.trace import TraceDataset


class TracePlot:
    """
    使用 Plotly Resampler 绘制大规模示波图的面板，仅保留曲线信息
    """

    def __init__(self):
        self._trace_dataset: TraceDataset | None = None
        self._traces: NDArray | None = None
        self._trace_names: list | None = None
        self._trace_name_prefix: list[str] | None = None
        self._fig = FigureWidgetResampler(go.Figure())
        self._fig.update_xaxes(  # type: ignore
            showline=True,
            linewidth=1,
            linecolor="#dddddd",
            mirror=True,
            gridwidth=1,
            gridcolor="#dddddd",
        )
        self._fig.update_yaxes(  # type: ignore
            showline=True, linewidth=1, linecolor="#dddddd", mirror=True, gridwidth=1, gridcolor="#dddddd"
        )
        self._fig.update_layout(
            plot_bgcolor="white",  # 绘图区背景白色
            paper_bgcolor="white",  # 整个画布背景白色
        )
        self._fig.update_layout(margin=dict(t=40, b=40, l=0))
        self._fig.update_layout(height=350)
        self._has_plot = False
        self._shift_map = {}

    def set_trace_dataset(self, trace_dataset: TraceDataset) -> None:
        """
        设置曲线
        :param trace_dataset: 曲线数据集
        :type trace_dataset: TraceDataset
        """
        self._trace_dataset = trace_dataset

    def plot_line(self, width=None, height=None) -> FigureWidgetResampler:
        """
        显示曲线， 默认不通过配置 show_trace 配置显示曲线时，则显示 通道 0 的所有曲线
        """
        if self._traces is None:
            self.show_trace()
        self._fig.data = []
        traces_list = [
            go.Scattergl(
                name=f"{self._trace_names[i]}",
                showlegend=True,
                line={"width": 1},
                mode="lines",
                y=self._shifted_trace(i),
            )
            for i in range(self._traces.shape[0])
        ]
        self._fig.add_traces(traces_list)
        if width is not None:
            self._fig.update_layout(width=width)
        if height is not None:
            self._fig.update_layout(height=height)
        self._fig.reload_data()
        self._has_plot = True
        return self._fig

    def _shifted_trace(self, i):
        trace = self._traces[i, :]
        if i in self._shift_map:
            shift = self._shift_map[i]
            return self._shift_array(trace, shift)
        else:
            return trace

    @staticmethod
    def _shift_array(a, shift, fill=0):
        result = np.full_like(a, fill, dtype=np.int16)
        if shift > 0:
            result[shift:] = a[:-shift]
        elif shift < 0:
            result[:shift] = a[-shift:]
        else:
            result[:] = a
        return result

    def shift(self, ch_idx: int, trace_idx: int, shift: int) -> None:
        """
        对指定通道和曲线进行偏移

        :param ch_idx: 通道索引
        :type ch_idx: int
        :param trace_idx: 曲线索引
        :type trace_idx: int
        :param shift: 偏移量
        :type shift: int
        """
        self._shift_map[trace_idx] = shift
        if self._has_plot:
            self.plot_line()

    def zoom(self, x: tuple[int, int] | None = None, y: tuple[int, int] | None = None) -> None:
        """
        曲线缩放
        """
        if x:
            self._fig.layout.xaxis.range = x
        if y:
            self._fig.layout.yaxis.range = y
        if not x and not y:
            self._fig.layout.xaxis.range = None
            self._fig.layout.yaxis.range = None

    def zoom_x(self, x: tuple[int, int]) -> None:
        """
        X 轴缩放
        """
        self.zoom(x=x, y=None)

    def zoom_y(self, y: tuple[int, int]) -> None:
        """
        Y 轴缩放
        """
        self.zoom(x=None, y=y)

    def show_trace(self, ch_idx: int | list[int] = 0, t_idx: list[slice, slice] | slice | None = None) -> None:
        """
        配置要显示的曲线

        :param ch_idx: 通道索引，可以是 int 或 list
        :type ch_idx: int | list[int]
        :param t_idx: 曲线切片，当传入切片列表时，列表与 ch_idx 列表匹配进行切片
        :type t_idx: list[slice, slice]
        """

        if isinstance(ch_idx, int):
            traces = self._trace_dataset.get_origin_data()[f"0/{ch_idx}/traces"]
            if t_idx is None:
                self._traces = traces
                self._trace_names = [f"T{i}" for i in range(traces.shape[0])]
            elif isinstance(t_idx, slice):
                self._traces = traces[t_idx]
                self._trace_names = [f"T{i}" for i in self._slice_to_list(t_idx, traces.shape[0])]
            elif isinstance(t_idx, list):
                if isinstance(t_idx, list):
                    _traces = []
                    _trace_names = []
                    for t_idx_slice in t_idx:
                        _traces.append(traces[t_idx_slice])
                        _trace_names.append([f"T{i}" for i in self._slice_to_list(t_idx_slice, traces.shape[0])])
                    self._traces = np.vstack(_traces)
                    self._trace_names = [x for row in _trace_names for x in row]
                else:
                    self._traces = traces[t_idx[0]]
                    self._trace_names = [f"T{i}" for i in self._slice_to_list(t_idx[0], traces.shape[0])]
        elif isinstance(ch_idx, list):
            traces_list = []
            trace_names_list = []
            for i, _ch_idx in enumerate(ch_idx):
                traces = self._trace_dataset.get_origin_data()[f"0/{_ch_idx}/traces"]
                name_ch_prefix = f"C{_ch_idx}-"
                if t_idx is None:
                    traces_list.append(traces)
                    trace_names_list.append([f"{name_ch_prefix}T{i}" for i in range(traces.shape[0])])
                elif isinstance(t_idx, slice):
                    traces_list.append(traces[t_idx])
                    trace_names_list.append(
                        [f"{name_ch_prefix}T{i}" for i in self._slice_to_list(t_idx, traces.shape[0])]
                    )
                elif isinstance(t_idx, list):
                    if len(ch_idx) != len(traces):
                        raise ValueError("通道数量和曲线切片数组长度不一致")
                    traces_list.append(traces[t_idx[i]])
                    trace_names_list.append(
                        [f"{name_ch_prefix}T{i}" for i in self._slice_to_list(t_idx[i], traces.shape[0])]
                    )
            self._traces = np.vstack(traces_list)
            self._trace_names = [name for sublist in trace_names_list for name in sublist]
        if self._has_plot:
            self.plot_line()

    @staticmethod
    def _slice_to_list(s: slice, length: int = None):
        """
        将 slice 对象展开为一个 list。

        参数:
            s (slice): 需要展开的切片对象
            length (int, 可选): 当 slice.stop 为 None 时，需要给定序列长度，用于确定范围。

        返回:
            list: 按 slice 展开的列表
        """
        start = s.start or 0
        step = s.step or 1

        if s.stop is None:
            if length is None:
                raise ValueError("slice.stop 为 None 时必须指定 length")
            stop = length
        else:
            stop = s.stop

        return list(range(start, stop, step))



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib

import traitlets
from anywidget import AnyWidget

from cracknuts import logger
from cracknuts.glitch.param_generator import VCCGlitchParamGenerator, GlitchGenerateParam


class GlitchTestPanel(AnyWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "GlitchTestPanelWidget.js"
    _css = ""

    tl_vcc_glitch_param_generator = traitlets.Dict({}).tag(sync=True)

    def __init__(self):
        super().__init__()
        self._logger = logger.get_logger(self)
        self._vcc_glitch_param_generator: VCCGlitchParamGenerator | None = None

    @traitlets.observe("vcc_glitch_param_generator")
    def on_vcc_glitch_param_change(self, change):
        vcc_glitch_param_generator_dict = change["new"]
        for param_type, param_dict in vcc_glitch_param_generator_dict.items():
            param = GlitchGenerateParam(**param_dict)
            param_type = param_type.lower()
            if param_type == "normal":
                self._vcc_glitch_param_generator.normal = param
            elif param_type == "glitch":
                self._vcc_glitch_param_generator.glitch = param
            elif param_type == "wait":
                self._vcc_glitch_param_generator.wait = param
            elif param_type == "count":
                self._vcc_glitch_param_generator.count = param
            elif param_type == "repeat":
                self._vcc_glitch_param_generator.repeat = param
            elif param_type == "interval":
                self._vcc_glitch_param_generator.interval = param



# Copyright 2024 CrackNuts. All rights reserved.

import abc
import itertools
from enum import Enum
import random


class GlitchGenerateParam:
    class Mode(Enum):
        INCREASE = 0
        DECREASE = 1
        RANDOM = 2
        FIXED = 3

    def __init__(self, mode, start, end, step, count):
        self.mode = mode
        self.start = start
        self.end = end
        self.step = step
        self.count = count

    def __str__(self):
        return str(self.__dict__)


class AbstractGlitchParamGenerator(abc.ABC):
    def __init__(self):
        self._generate_params = self._build()
        self._current_index = 0

    def _build(self) -> list[GlitchGenerateParam]: ...

    @staticmethod
    def _do_build(param: GlitchGenerateParam):
        if param.mode == GlitchGenerateParam.Mode.FIXED:
            return [param.start for _ in range(param.count)]
        if param.mode == GlitchGenerateParam.Mode.INCREASE:
            return list(AbstractGlitchParamGenerator._frange(param.start, param.end, param.step)) * param.count
        if param.mode == GlitchGenerateParam.Mode.DECREASE:
            return list(AbstractGlitchParamGenerator._frange(param.end, param.start, -param.step)) * param.count
        if param.mode == GlitchGenerateParam.Mode.RANDOM:
            return [random.randint(param.start, param.end) for _ in range(param.count)]
        return []

    @staticmethod
    def _frange(start, stop, step):
        x = start
        while x <= stop:
            yield round(x, 10)
            x += step

    def total(self):
        return len(self._generate_params)

    def next(self):
        param = self._generate_params[self._current_index]
        self._current_index += 1
        if self._current_index >= len(self._generate_params):
            self._current_index = 0
        return param


class VCCGlitchParam:
    def __init__(self, normal: int, wait: int, glitch: int, count: int, repeat: int, interval: int):
        self.normal = normal
        self.wait = wait
        self.glitch = glitch
        self.count = count
        self.repeat = repeat
        self.interval = interval


class VCCGlitchParamGenerator(AbstractGlitchParamGenerator):
    def __init__(
        self,
        normal: GlitchGenerateParam,
        wait: GlitchGenerateParam,
        glitch: GlitchGenerateParam,
        count: GlitchGenerateParam,
        repeat: GlitchGenerateParam,
        interval: GlitchGenerateParam,
    ):
        self.normal = normal
        self.wait = wait
        self.glitch = glitch
        self.count = count
        self.repeat = repeat
        self.interval = interval
        super().__init__()

    def _build(self):
        normal_list = []
        wait_list = []
        glitch_list = []
        count_list = []
        repeat_list = []
        interval_list = []
        if self.normal is not None:
            normal_list = self._do_build(self.normal)
        if self.wait is not None:
            wait_list = self._do_build(self.wait)
        if self.glitch is not None:
            glitch_list = self._do_build(self.glitch)
        if self.count is not None:
            count_list = self._do_build(self.count)
        if self.repeat is not None:
            repeat_list = self._do_build(self.repeat)
        if self.interval is not None:
            interval_list = self._do_build(self.interval)

        return [
            VCCGlitchParam(*combo)
            for combo in itertools.product(
                normal_list,
                wait_list,
                glitch_list,
                count_list,
                repeat_list,
                interval_list,
            )
        ]

    def __str__(self):
        return str(self.__dict__)


class GNDGlitchParamGenerator(VCCGlitchParamGenerator): ...



# Copyright 2024 CrackNuts. All rights reserved.

from numba import prange, njit
from numpy.typing import NDArray
import numpy as np


@njit(parallel=True, cache=True)
def minmax(value: NDArray[np.int16], mn: int, mx: int, down_count: int) -> tuple[NDArray[np.int32], NDArray[np.int16]]:
    mn = max(0, mn)
    mx = min(value.shape[0], mx)
    ds = max(1, int((mx - mn) / down_count))
    if ds == 1:
        _index = np.arange(mn, mx, dtype=np.int32)
        _value = value[mn:mx]
        return _index, _value
    sample_count = (mx - mn) // ds
    down_index = np.empty(sample_count * 2, dtype=np.int32)
    down_value = np.empty(sample_count * 2, dtype=np.int16)

    for i in prange(sample_count):
        start = mn + i * ds
        end = start + ds
        block = value[start:end]

        down_index[2 * i] = start
        down_index[2 * i + 1] = start
        down_value[2 * i] = block.max()
        down_value[2 * i + 1] = block.min()
    return down_index, down_value



# Copyright 2024 CrackNuts. All rights reserved.

from cracknuts.acquisition.glitch_acquisition import GlitchAcquisition
from cracknuts.cracker.cracker_basic import CrackerBasic
from cracknuts.acquisition import Acquisition

from cracknuts.jupyter.workbench_g1_panel import WorkbenchG1Panel
from cracknuts.jupyter.acquisition_panel import AcquisitionPanelWidget
from cracknuts.jupyter.cracker_s1_panel import CrackerPanelWidget
from cracknuts.jupyter.cracknuts_panel import CracknutsPanelWidget
from cracknuts.jupyter.trace_panel import TracePanelWidget
from cracknuts.jupyter.scope_panel import ScopePanelWidget
from cracknuts.trace.trace import TraceDataset
from cracknuts import logger

_logger = logger.get_logger("cracknuts.jupyter")


def show_panel(acq: Acquisition|GlitchAcquisition):
    if isinstance(acq, GlitchAcquisition):
        return WorkbenchG1Panel(cracker=acq.cracker, acquisition=acq)
    elif isinstance(acq, Acquisition):
        return CracknutsPanelWidget(acquisition=acq)
    else:
        _logger.error(f"Unsupported acquisition type: {acq}")
        return None


def display_cracknuts_panel(acq: "Acquisition"):
    acq.sample_length = -1  # ignore the sample_length of acquisition when use ui.
    cnpw = CracknutsPanelWidget(acquisition=acq)
    # cnpw.sync_config()
    # cnpw.listen_config()
    return cnpw


def display_trace_panel(trace_dataset: TraceDataset = None):
    tpw = TracePanelWidget()
    if trace_dataset:
        tpw.set_trace_dataset(trace_dataset)
    return tpw


def display_scope_panel(acq: "Acquisition"):
    acq.sample_length = -1  # ignore the sample_length of acquisition when use ui.
    return ScopePanelWidget(acquisition=acq)


def display_acquisition_panel(acq: "Acquisition"):
    acq.sample_length = -1  # ignore the sample_length of acquisition when use ui.
    acqw = AcquisitionPanelWidget(acquisition=acq)
    acqw.sync_config_from_acquisition()
    return acqw


def display_cracker_panel(cracker: "CrackerBasic"):
    cpw = CrackerPanelWidget(cracker=cracker)
    cpw.read_config_from_cracker()
    cpw.listen_cracker_config()
    return cpw


__all__ = [
    "show_panel"
]



# Copyright 2024 CrackNuts. All rights reserved.

__version__ = "0.20.0-beta.0"

import sys

from cracknuts import logger
from cracknuts.jupyter import show_panel
from cracknuts.cracker import cracker_s1, cracker_g1
from cracknuts.acquisition import simple_acq, simple_glitch_acq


_logger = logger.get_logger("cracknuts")

try:
    from IPython.display import display as _ipython_display

    if "ipykernel" not in sys.modules:
        _ipython_display = None
except ImportError as e:
    _logger.error(f"ImportError: {e.args}")
    _ipython_display = None


def version():
    return __version__


__all__ = [
    "cracker_s1",
    "cracker_g1",
    "simple_acq",
    "simple_glitch_acq",
    "show_panel",
    "version",
]



# Copyright 2024 CrackNuts. All rights reserved.

import typing
from typing import Any

from traitlets import traitlets

from cracknuts.cracker.cracker_g1 import ConfigG1, CrackerG1
from cracknuts.jupyter.config.config_glitch import ConfigG1Glitch
from cracknuts.jupyter.cracker_s1_panel import CrackerPanelWidget
from cracknuts.jupyter.ui_sync import observe_interceptor


class CrackerG1Panel(CrackerPanelWidget, ConfigG1Glitch):

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.cracker: CrackerG1 = self.cracker

    def get_cracker_panel_config(self) -> ConfigG1:
        """
        Get the current configuration from panel(Jupyter widget UI)

        The nut_glitch_vcc_normal is mapped to nut_voltage,
        and nut_glitch_clock_enable is mapped to nut_clock_enable.

        """
        s1_panel_config = super().get_cracker_panel_config()
        panel_config = ConfigG1()
        for k, v in s1_panel_config.__dict__.items():
            setattr(panel_config, k, v)
        panel_config.glitch_vcc_normal = self.nut_voltage  # use nut_voltage as glitch_vcc_normal
        panel_config.glitch_vcc_config_wait = self.glitch_vcc_config_wait
        panel_config.glitch_vcc_config_level = self.glitch_vcc_config_level
        panel_config.glitch_vcc_config_count = self.glitch_vcc_config_count
        panel_config.glitch_vcc_config_delay = self.glitch_vcc_config_delay
        panel_config.glitch_vcc_config_repeat = self.glitch_vcc_config_repeat

        panel_config.glitch_gnd_normal = self.glitch_gnd_normal
        panel_config.glitch_gnd_config_wait = self.glitch_gnd_config_wait
        panel_config.glitch_gnd_config_level = self.glitch_gnd_config_level
        panel_config.glitch_gnd_config_count = self.glitch_gnd_config_count
        panel_config.glitch_gnd_config_delay = self.glitch_gnd_config_delay
        panel_config.glitch_gnd_config_repeat = self.glitch_gnd_config_repeat

        # Use glitch_clock_normal_rate as nut_clock. This field is ignored in the Jupyter UI,
        # but in a pure Python environment, users can read the nut clock configuration from it.
        panel_config.nut_clock = self.get_glitch_clock_normal_freq()
        panel_config.glitch_clock_len_normal = len(self.glitch_clock_wave_normal)
        panel_config.glitch_clock_wave_normal = self.glitch_clock_wave_normal
        panel_config.glitch_clock_config_len_glitch = len(self.glitch_clock_config_wave_glitch)
        panel_config.glitch_clock_config_wave_glitch = self.glitch_clock_config_wave_glitch
        panel_config.glitch_clock_config_wait = self.glitch_clock_config_wait
        panel_config.glitch_clock_config_delay = self.glitch_clock_config_delay
        panel_config.glitch_clock_config_repeat = self.glitch_clock_config_repeat
        panel_config.glitch_clock_enable = self.nut_clock_enable # use nut_clock_enable as glitch_clock_enable

        return panel_config

    def _panel_config_item_process(self, name: str, value: object) -> object:
        """
        Process a single panel configuration item:
        return the corresponding value when no extra processing is needed,
        or return None to indicate that the item is skipped or already processed by this function.
        """
        if name == "nut_voltage":
            return None # ignore, and use glitch_vcc_normal instead
        elif name == "nut_clock":
            return None # ignore, and use glitch_clock_normal_rate instead
        elif name == "nut_clock_enable":
            return None # ignore, and use glitch_clock_enable instead
        elif name == "glitch_vcc_normal":
            self.glitch_vcc_normal = value
            self.nut_voltage = value
            return None
        elif name == "glitch_clock_wave_normal":
            self.glitch_clock_wave_normal = typing.cast(list[float], value)
            self.nut_clock = self.get_glitch_clock_normal_freq()
            self.glitch_clock_rate = f"{round(self.nut_clock / 1000, 2)}M"
        elif name == "glitch_clock_enable":
            self.glitch_clock_enable = typing.cast(bool, value)
            self.nut_clock_enable = self.glitch_clock_enable
        elif name == "glitch_clock_len_normal":
            return None # ignore, already processed by glitch_clock_wave_normal
        elif name == "glitch_clock_config_len_glitch":
            return None # ignore, already processed by glitch_clock_wave_glitch
        elif name == "glitch_clock_config_wave_glitch":
            # 忽略，因为这里的波形配置是通过长度计算来的失真数据
            return None
        elif name in ("glitch_vcc_arm", "glitch_gnd_arm", "glitch_clock_arm"):
            return None # ignore these arm settings in G1 panel
        else:
            value = super()._panel_config_item_process(name, value)
        return value

    @traitlets.observe("nut_voltage")
    @observe_interceptor
    def nut_voltage_changed(self, change):
        self.cracker.nut_voltage(change.get("new"))

    @traitlets.observe("nut_clock")
    @observe_interceptor
    def nut_clock_changed(self, change):
        super().nut_clock_changed(change)

    @traitlets.observe("glitch_vcc_normal")
    @observe_interceptor
    def glitch_vcc_normal_changed(self, change):
        self.cracker.glitch_vcc_normal(change.get("new"))

    # TODO 这里有个同步的问题，就是电压和时钟实际上在界面中有两个位置（基本设置和故障注入），需要同步更新


# Copyright 2024 CrackNuts. All rights reserved.

# Copyright 2024 CrackNuts. All rights reserved.
import os
import re
import struct
import typing

import pandas as pd
from scipy.interpolate import interp1d
import importlib.util
from cracknuts.cracker import protocol
from cracknuts.cracker.cracker_basic import ConfigBasic, connection_status_check
from cracknuts.cracker.cracker_s1 import ConfigS1, CrackerS1
from cracknuts.cracker.protocol import Command


wave_80m = [
    3.2,
    0.0
]
wave_40m = [
    3.2, 3.2,
    0.0, 0.0
]
wave_20m = [
    3.2, 3.2, 3.2, 3.2,
    0.0, 0.0, 0.0, 0.0,
]
wave_16m = [
    3.2, 3.2, 3.2, 3.2, 3.2,
    0.0, 0.0, 0.0, 0.0, 0.0
]
wave_10m = [
    3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2,
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
]
wave_8m = [
    3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2,
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
]
wave_4m = [
    3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2, 3.2,
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
]


class ConfigG1(ConfigS1):
    def __init__(self):
        super().__init__()
        self.glitch_vcc_arm = False
        self.glitch_vcc_normal = 3.3
        self.glitch_vcc_config_wait = 1
        self.glitch_vcc_config_level = self.glitch_vcc_normal
        self.glitch_vcc_config_count = 1
        self.glitch_vcc_config_delay = 0
        self.glitch_vcc_config_repeat = 1

        self.glitch_gnd_arm = False
        self.glitch_gnd_normal = 0
        self.glitch_gnd_config_wait = 1
        self.glitch_gnd_config_level = self.glitch_gnd_normal
        self.glitch_gnd_config_count = 1
        self.glitch_gnd_config_delay = 0
        self.glitch_gnd_config_repeat = 1

        self.glitch_clock_arm: bool = False
        self.glitch_clock_len_normal: int = len(wave_8m)
        self.glitch_clock_wave_normal: list[float] = wave_8m # 默认时钟8mhz
        self.glitch_clock_config_len_glitch: int = len(wave_4m)
        self.glitch_clock_config_wave_glitch: list[float] = wave_4m # 默认glitch示例时钟4mhz
        self.glitch_clock_config_wait: int = 1
        self.glitch_clock_config_delay: int = 1
        self.glitch_clock_config_repeat: int = 1
        self.glitch_clock_enable: bool = True

class CommandG1(Command):
    GLITCH_VCC_ARM = 0x0310
    GLITCH_VCC_RESET = 0x0311
    GLITCH_VCC_FORCE = 0x0312
    GLITCH_VCC_CONFIG = 0x0313
    GLITCH_VCC_NORMAL = 0x0314

    GLITCH_GND_ARM = 0x0320
    GLITCH_GND_RESET = 0x0321
    GLITCH_GND_FORCE = 0x0322
    GLITCH_GND_CONFIG = 0x0323
    GLITCH_GND_NORMAL = 0x0324

    GLITCH_CLK_ARM = 0x0330
    GLITCH_CLK_RESET = 0x0331
    GLITCH_CLK_FORCE = 0x0332
    GLITCH_CLK_CONFIG = 0x0333
    GLITCH_CLK_LEN_NORMAL = 0x0334
    GLITCH_CLK_WAVE_NORMAL = 0x0335
    GLITCH_CLK_LEN_GLITCH = 0x0336
    GLITCH_CLK_WAVE_GLITCH = 0x0337


def _build_interp_func(map_path) -> typing.Tuple[interp1d, interp1d]:
    df = pd.read_csv(map_path)
    x = df["code"].map(lambda v: int(v, 16)).values
    y = df["voltage"].values
    return interp1d(x, y, kind="linear", fill_value="extrapolate"), interp1d(y, x, kind="linear", fill_value="extrapolate")


_voltage_map_dir = os.path.join(os.path.dirname(importlib.util.find_spec("cracknuts").origin), "glitch", "voltage_map")
_vcc_interp_func_cv, _vcc_interp_func_vc = _build_interp_func(os.path.join(_voltage_map_dir, "vcc.csv"))
_gnd_interp_func_cv, _gnd_interp_func_vc = _build_interp_func(os.path.join(_voltage_map_dir, "gnd.csv"))
_clk_interp_func_cv, _clk_interp_func_vc = _build_interp_func(os.path.join(_voltage_map_dir, "clk.csv"))

SYS_CLK_PERIOD_S = 6.25e-9  # g1 sys clock period in seconds

def get_clock_from_wave_length(wave_len: int) -> float:
    return 1 / (wave_len * SYS_CLK_PERIOD_S) / 1000


class CrackerG1(CrackerS1):
    def __init__(
        self,
        address: tuple | str | None = None,
        bin_server_path: str | None = None,
        bin_bitstream_path: str | None = None,
        operator_port: int = None,
    ):
        super().__init__(address, bin_server_path, bin_bitstream_path, operator_port)
        self._config: ConfigG1 = self._config

    @staticmethod
    def _get_dac_code_from_voltage(voltage: float, interp_func: interp1d) -> int:
        dac_code = int(round(interp_func(voltage * 1000).item()))
        return 0 if dac_code < 0 else 0x03FF if dac_code > 0x03FF else dac_code

    @staticmethod
    def _get_voltage_from_dac_code(dac_code: int, interp_func: interp1d) -> float:
        voltage = interp_func(dac_code).item() / 1000
        return voltage

    def glitch_vcc_arm(self):
        """
        设置glitch VCC 为armed状态。
        """
        self._glitch_vcc_arm(True)

    # def glitch_vcc_disable(self):
    #     self._glitch_vcc_enable(False)

    def _glitch_vcc_arm(self, enable: bool):
        payload = struct.pack(">?", enable)
        self._logger.debug(f"glitch_vcc_enable payload: {payload.hex()}")
        status, res = self.send_with_command(CommandG1.GLITCH_VCC_ARM, payload=payload)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def glitch_vcc_config(self, wait: int, level: int | float | str, length: int, delay: int, repeat: int):
        """
        配置glitch ::

            v_normal────┬───────────────────┐       ┌────────────┐       ┌───────
                        |       wait        | count |    delay   | count |
                        |                   |       |            |       |
                        |                   └───────┘            └───────┘ <--------- level voltage
                        |                       └────────────────────┘
                         Trigger                         repeat

        :param wait: glitch产生前等待时间（时钟个数）
        :type wait: int
        :param level: Glitch DAC电压值
        :type level: int
        :param length: Glitch持续时间, 单位 10ns
        :type delay: int
        :param repeat: Glitch重复次数
        :type repeat: int
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        level = self._parse_voltage(level)
        level_code = self._get_dac_code_from_voltage(level, _vcc_interp_func_vc)
        payload = struct.pack(">IIIII", wait, level_code, length, delay, repeat)
        self._logger.debug(f"glitch_vcc_config payload: {payload.hex()}")
        status, res = self.send_with_command(CommandG1.GLITCH_VCC_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.glitch_vcc_config_wait = wait
            self._config.glitch_vcc_config_level = level
            self._config.glitch_vcc_config_count = length
            self._config.glitch_vcc_config_delay = delay
            self._config.glitch_vcc_config_repeat = repeat
        return status, res

    def glitch_vcc_reset(self):
        """
        重置glitch VCC配置。
        """
        self._logger.debug(f"glitch_vcc_reset payload: {None}")
        status, res = self.send_with_command(CommandG1.GLITCH_VCC_RESET, payload=None)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def glitch_vcc_force(self):
        """
        强制触发glitch VCC。
        """
        self._logger.debug(f"glitch_vcc_force payload: {None}")
        status, res = self.send_with_command(CommandG1.GLITCH_VCC_FORCE, payload=None)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def nut_voltage(self, voltage: float | str | int) -> tuple[int, None]:
        return self.glitch_vcc_normal(voltage)

    def glitch_vcc_normal(self, voltage: float | str | int) -> tuple[int, None]:
        """
        设置glitch VCC正常电压值。
        """
        voltage = self._parse_voltage(voltage)
        dac_code = self._get_dac_code_from_voltage(voltage, _vcc_interp_func_vc)
        payload = struct.pack(">I", dac_code)
        self._logger.debug(f"glitch_vcc_normal payload: {payload}")
        status, res = self.send_with_command(CommandG1.GLITCH_VCC_NORMAL, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.glitch_vcc_normal = voltage
            self._config.nut_voltage = voltage
        return status, res

    def _parse_voltage(self, voltage: float | str | int):
        if isinstance(voltage, str):
            m = re.match(r"^(\d+(?:\.\d+)?)([mM]?[vV])?$", voltage)
            if not m:
                self._logger.error(
                    "Input format error: " "it should be a number or a string and end with a unit in V or mV."
                )
                return self.NON_PROTOCOL_ERROR, None
            voltage = float(m.group(1))
            if m.group(2):
                unit = m.group(2)
            else:
                unit = "v"
            if unit.lower() == "mv":
                voltage = voltage / 1000
        else:
            voltage = float(voltage)
        return voltage

    def glitch_gnd_arm(self):
        self._glitch_gnd_arm(True)

    # def glitch_gnd_disable(self):
    #     self._glitch_gnd_arm(False)

    def _glitch_gnd_arm(self, enable: bool):
        payload = struct.pack(">?", enable)
        self._logger.debug(f"glitch_gnd_enable payload: {payload.hex()}")
        status, res = self.send_with_command(CommandG1.GLITCH_GND_ARM, payload=payload)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def glitch_gnd_config(self, wait: int, level: int, length: int, delay: int, repeat: int):
        level = self._parse_voltage(level)
        level_code = self._get_dac_code_from_voltage(level, _gnd_interp_func_vc)
        payload = struct.pack(">IIIII", wait, level_code, length, delay, repeat)
        self._logger.debug(f"glitch_gnd_config payload: {payload.hex()}")
        status, res = self.send_with_command(CommandG1.GLITCH_GND_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.glitch_gnd_config_wait = wait
            self._config.glitch_gnd_config_level = level
            self._config.glitch_gnd_config_count = length
            self._config.glitch_gnd_config_delay = delay
            self._config.glitch_gnd_config_repeat = repeat
        return status, res

    def glitch_gnd_reset(self):
        self._logger.debug(f"glitch_vcc_reset payload: {None}")
        status, res = self.send_with_command(CommandG1.GLITCH_GND_RESET, payload=None)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def glitch_gnd_force(self):
        self._logger.debug(f"glitch_vcc_force payload: {None}")
        status, res = self.send_with_command(CommandG1.GLITCH_GND_FORCE, payload=None)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def glitch_gnd_normal(self, voltage: float | str | int) -> tuple[int, None]:
        dac_code = self._get_dac_code_from_voltage(self._parse_voltage(voltage), _gnd_interp_func_vc)
        payload = struct.pack(">I", dac_code)
        self._logger.debug(f"glitch_gnd_normal payload: {payload}")
        status, res = self.send_with_command(CommandG1.GLITCH_GND_NORMAL, payload=payload)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def _get_config_bytes_format(self) -> tuple[dict[str, str], ConfigBasic]:
        bytes_format, config = super()._get_config_bytes_format()
        bytes_format.update(
            {
                "glitch_vcc_arm": "?",
                "glitch_vcc_normal": "I",
                "glitch_vcc_config_wait": "I",
                "glitch_vcc_config_level": "I",
                "glitch_vcc_config_count": "I",
                "glitch_vcc_config_delay": "I",
                "glitch_vcc_config_repeat": "I",
                "glitch_gnd_arm": "?",
                "glitch_gnd_normal": "I",
                "glitch_gnd_config_wait": "I",
                "glitch_gnd_config_level": "I",
                "glitch_gnd_config_count": "I",
                "glitch_gnd_config_delay": "I",
                "glitch_gnd_config_repeat": "I",
                "glitch_clock_arm": "?",
                "glitch_clock_len_normal": "I",
                "glitch_clock_wave_normal_fake": "I",  # todo 这里无效，后续在server端删除
                "glitch_clock_config_len_glitch": "I",
                "glitch_clock_config_wave_glitch_fake": "I",  # todo 这里无效，后续在server端删除
                "glitch_clock_config_wait": "I", # todo 这里临时注释，后续在server端添加
                "glitch_clock_config_delay": "I",
                "glitch_clock_config_repeat": "I",
                # "glitch_clock_enable": "?",  # todo 这里临时注释，后续在server端添加
            }
        )
        config = ConfigG1()
        return bytes_format, config

    def _parse_config_special_case(self, k, v):
        if k in ("glitch_vcc_normal", "glitch_vcc_config_level"):
            v = round(self._get_voltage_from_dac_code(v, _vcc_interp_func_cv), 1)
        elif k in ("glitch_gnd_normal", "glitch_gnd_config_level"):
            v = round(self._get_voltage_from_dac_code(v, _gnd_interp_func_cv), 1)
        return v

    @connection_status_check
    def get_current_config(self) -> ConfigG1 | None:
        cur_cfg = typing.cast(ConfigG1, super().get_current_config())
        # vcc
        cur_cfg.nut_voltage = cur_cfg.glitch_vcc_normal
        # clk normal
        clk_rate, clk_wave = self._get_clock_from_wave_len(cur_cfg.glitch_clock_len_normal)
        cur_cfg.nut_clock = round(clk_rate)
        cur_cfg.glitch_clock_wave_normal = clk_wave
        # clk normal enable
        cur_cfg.nut_clock_enable = self._get_cracker_nut_clock_enabled()
        cur_cfg.glitch_clock_enable = cur_cfg.nut_clock_enable
        # clk glitch
        glitch_clk_rate, glitch_clk_wave = self._get_clock_from_wave_len(cur_cfg.glitch_clock_config_len_glitch)
        cur_cfg.glitch_clock_config_wave_glitch = glitch_clk_wave
        return cur_cfg

    # def _get_glitch_normal_clock(self) -> tuple[float, list[float]]:
    #     s, r = self.register_read(base_address=0x43c10000, offset=0x1D0)
    #
    #     if s != protocol.STATUS_OK:
    #         return 0.0, []
    #     try:
    #         wave_len = struct.unpack('>I', r)[0]
    #         return self._get_clock_from_wave_len(wave_len)
    #     except struct.error as e:
    #         self._logger.error(f"Unpack nut clock failed: {e}")
    #         return 0.0, []
    #
    # def _get_glitch_clock(self) -> tuple[float, list[float]]:
    #     s, r = self.register_read(base_address=0x43c10000, offset=0x1DC)
    #
    #     if s != protocol.STATUS_OK:
    #         return 0.0, []
    #     try:
    #         wave_len = struct.unpack('>I', r)[0]
    #         return self._get_clock_from_wave_len(wave_len)
    #     except struct.error as e:
    #         self._logger.error(f"Unpack nut clock failed: {e}")
    #         return 0.0, []

    @staticmethod
    def _get_clock_from_wave_len(wave_len: int) -> tuple[float, list[float]]:
        """
        获取cracker当前nut时钟频率，单位kHz。
        :return: nut时钟频率，单位kHz
        :rtype: float
        """
        # 这里在cracker中只能拿到寄存器中的波形长度信息，没有波形信息，所有只根据对半将波形分为高低电平来计算频率
        return get_clock_from_wave_length(wave_len), [3.2 if i < wave_len / 2 else 0.0 for i in range(wave_len)]

    def _get_cracker_nut_vcc_voltage(self) -> float:
        """
        获取cracker当前nut电压，单位V。
        :return: nut电压，单位V
        :rtype: float
        """
        s, r = self.register_read(base_address=0x43c10000, offset=0x150)
        if s != protocol.STATUS_OK:
            return 0.0
        try:
            dac_code = struct.unpack('>I', r)[0]
            return self._get_voltage_from_dac_code(dac_code, _vcc_interp_func_cv)
        except struct.error as e:
            self._logger.error(f"Unpack nut voltage failed: {e}")
            return 0.0

    def _get_cracker_nut_gnd_voltage(self) -> float:
        """
        获取cracker当前nut电压，单位V。
        :return: nut电压，单位V
        :rtype: float
        """
        s, r = self.register_read(base_address=0x43c10000, offset=0x190)
        if s != protocol.STATUS_OK:
            return 0.0
        try:
            dac_code = struct.unpack('>I', r)[0]
            return self._get_voltage_from_dac_code(dac_code, _vcc_interp_func_cv)
        except struct.error as e:
            self._logger.error(f"Unpack nut voltage failed: {e}")
            return 0.0

    def _get_cracker_nut_clock_enabled(self) -> bool:
        """
        获取cracker当前nut时钟是否enabled。
        :return: nut时钟是否enabled
        :rtype: bool
        """
        s, r = self.register_read(base_address=0x43c10000, offset=0x1C4)
        if s != protocol.STATUS_OK:
            return False
        try:
            enabled = struct.unpack('>I', r)[0]
            return enabled == 0
        except struct.error as e:
            self._logger.error(f"Unpack nut clock enabled failed: {e}")
            return False

    def get_default_config(self) -> ConfigS1:
        return ConfigG1()

    def write_config_to_cracker(self, config: ConfigG1):
        self._osc_set_analog_all_channel_enable({0: config.osc_channel_0_enable, 1: config.osc_channel_1_enable})
        self.osc_analog_gain(0, config.osc_channel_0_gain)
        self.osc_analog_gain(1, config.osc_channel_1_gain)
        self.osc_sample_length(config.osc_sample_length)
        self.osc_sample_delay(config.osc_sample_delay)
        self.osc_sample_clock(config.osc_sample_clock)
        self.osc_sample_clock_phase(config.osc_sample_phase)
        self.osc_trigger_source(config.osc_trigger_source)
        self.osc_trigger_mode(config.osc_trigger_mode)
        self.osc_trigger_edge(config.osc_trigger_edge)
        self.osc_trigger_level(config.osc_trigger_edge_level)

        self.spi_config(
            config.nut_spi_speed,
            config.nut_spi_cpol,
            config.nut_spi_cpha,
            config.nut_spi_csn_auto,
            config.nut_spi_csn_delay,
        )
        self._spi_enable(config.nut_spi_enable)

        self.uart_config(
            config.nut_uart_baudrate,
            config.nut_uart_bytesize,
            config.nut_uart_parity,
            config.nut_uart_stopbits,
        )
        self._uart_io_enable(config.nut_uart_enable)

        self.i2c_config(config.nut_i2c_dev_addr, config.nut_i2c_speed)
        self._i2c_enable(config.nut_i2c_enable)

        ## Clock glitching settings ...
        # 要先设置时钟（或者说是 glitch clock normal ）
        # 然后设置电压（或者说是 glitch vcc normal）因为nut需要先有时钟后上电才能正常工作
        self.glitch_clock_normal(config.glitch_clock_wave_normal)
        self.glitch_clock_config(
            config.glitch_clock_config_wave_glitch,
            config.glitch_clock_config_wait,
            config.glitch_clock_config_repeat,
            config.glitch_clock_config_delay
        )
        self._nut_set_clock_enable(config.nut_clock_enable)

        self.glitch_vcc_normal(config.glitch_vcc_normal)
        self.glitch_vcc_config(
            config.glitch_gnd_config_wait,
            config.glitch_vcc_config_level,
            config.glitch_vcc_config_count,
            config.glitch_vcc_config_delay,
            config.glitch_vcc_config_repeat,
        )

        self.glitch_gnd_normal(config.glitch_gnd_normal)
        self.glitch_gnd_config(
            config.glitch_gnd_config_wait,
            config.glitch_gnd_config_level,
            config.glitch_gnd_config_count,
            config.glitch_gnd_config_delay,
            config.glitch_gnd_config_repeat,
        )

        self._nut_set_enable(config.nut_enable)


    # def set_glitch_test_params(self, param):
    #     self._glitch_test_params = param
    #
    # def get_glitch_test_params(self):
    #     return self._glitch_test_params

    def glitch_clock_normal(self, wave: list[float]) -> tuple[int, None|bytes]:
        status, res = self.register_write(base_address=0x43c10000, offset=0x1D0, data=len(wave))
        if status != protocol.STATUS_OK:
            return status, res
        for voltage in wave:
            status, res = self.register_write(base_address=0x43c10000, offset=0x1D4, data=self._get_dac_code_from_voltage(voltage, _clk_interp_func_vc))
            if status != protocol.STATUS_OK:
                return status, res
        self._config.glitch_clock_wave_normal = wave
        return protocol.STATUS_OK, None

    def _glitch_clock_normal_enable(self, enable: bool):
        return self.register_write(base_address=0x43c10000, offset=0x1C4, data=0 if enable else 1)

    def glitch_clock_normal_enable(self):
        return self._glitch_clock_normal_enable(True)

    def glitch_clock_normal_disable(self):
        return self._glitch_clock_normal_enable(False)

    def glitch_clock_config(self, wave: list[float], wait: int, delay: int, repeat: int) -> tuple[int, None|bytes]:
        status, res = self.register_write(base_address=0x43c10000, offset=0x1DC, data=len(wave))
        if status != protocol.STATUS_OK:
            return status, res
        for voltage in wave:
            status, res = self.register_write(base_address=0x43c10000, offset=0x1E0, data=self._get_dac_code_from_voltage(voltage, _clk_interp_func_vc))
            if status != protocol.STATUS_OK:
                return status, res
        status, res = self.register_write(base_address=0x43c10000, offset=0x1D8, data=wait)
        if status != protocol.STATUS_OK:
            return status, res
        self.register_write(base_address=0x43c10000, offset=0x1E4, data=delay)
        if status != protocol.STATUS_OK:
            return status, res
        self.register_write(base_address=0x43c10000, offset=0x1E8, data=repeat)
        if status != protocol.STATUS_OK:
            return status, res

        self._config.glitch_clock_config_wave_glitch = wave
        self._config.glitch_clock_config_len_glitch = len(wave)
        self._config.glitch_clock_config_wait = wait
        self._config.glitch_clock_config_repeat = repeat
        self._config.glitch_clock_config_delay = delay

        return protocol.STATUS_OK, None

    def glitch_clock_arm(self):
        return self.register_write(base_address=0x43c10000, offset=0X1C8, data=1)

    def glitch_clock_reset(self):
        return self._glitch_clock_normal_enable(False)

    def glitch_clock_force(self):
        return self.register_write(base_address=0x43c10000, offset=0x1CC, data=1)

    def _nut_set_clock_enable(self, enable: bool) -> tuple[int, None]:
        status, res = self._glitch_clock_normal_enable(enable)
        if status == protocol.STATUS_OK:
            self._config.nut_clock_enable = enable
        return status, res

    def nut_clock_freq(self, clock: int | str) -> tuple[int, None]:
        """
        Set nut clock.

        :param clock: Clock frequency in kHz or string with unit 'M', e.g., '8M' for 8000 kHz
        :type clock: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(clock, str):
            m = re.match(r"^(\d+)[mM]$", clock)
            if m:
                clock = int(m.group(1)) * 1000
            else:
                self._logger.error("Input format error: it should be an integer or a string end with 'M'.")
                return protocol.STATUS_ERROR, None
        if clock == 80000:
            wave = wave_80m
        elif clock == 40000:
            wave = wave_40m
        elif clock == 20000:
            wave = wave_20m
        elif clock == 16000:
            wave = wave_16m
        elif clock == 10000:
            wave = wave_10m
        elif clock == 8000:
            wave = wave_8m
        elif clock == 4000:
            wave = wave_4m
        else:
            self._logger.error(f"Unknown clock type: {clock}, only support 80M, 40M, 20M, 16M, 10M, 8M, 4M")
            return protocol.STATUS_ERROR, None
        status, res = self.glitch_clock_normal(wave)
        if status == protocol.STATUS_OK:
            self._config.nut_clock = clock
        return status, res



# Copyright 2024 CrackNuts. All rights reserved.

import typing

from cracknuts.acquisition.acquisition import Acquisition
from cracknuts.acquisition.glitch_acquisition import GlitchAcquisition
from cracknuts.cracker.cracker_s1 import CrackerS1
from cracknuts.cracker.cracker_g1 import CrackerG1


def simple_acq(
        cracker: CrackerS1,
        init_func: typing.Callable[[CrackerS1], None] = lambda cracker: None,
        do_func: typing.Callable[[CrackerS1, int], dict[str, bytes]] = lambda cracker, count: {},
        finish_func: typing.Callable[[CrackerS1], None] = lambda cracker: None,
        **kwargs,
) -> 'Acquisition':
    """
    创建一个简单的Acquisition子类实例。

    :param cracker: Cracker实例
    :type cracker: CrackerS1
    :param init_func:
        初始化函数，该函数接受一个CrackerS1实例作为参数，用于在采集开始前进行初始化操作。
        参数说明：
            - cracker: (CrackerS1) Cracker实例
            - 返回值：无
        示例：
            cracker.nut_voltage_enable()
            cracker.nut_voltage(3.3)
            cracker.nut_clock_enable()
            cracker.nut_clock_freq('8M')
            cracker.uart_io_enable()
            status, ret = cracker.uart_transmit_receive(cmd_set_aes_enc_key + aes_key, timeout=1000, rx_count=6)
    :type init_func: typing.Callable[[CrackerS1], None]
    :param do_func:
        执行函数，具体的采集逻辑实现，该函数接受一个CrackerS1实例和当前采集计数作为参数，返回一个包含采集数据的字典。
        参数说明：
            - cracker: (CrackerS1) Cracker实例
            - count: (int) 当前采集计数，从0开始
            - 返回值：dict[str, bytes]，包含采集数据的字典，格式如下：
              {
                  "plaintext": 明文数据的字节串,
                  "ciphertext": 密文数据的字节串,
                  "key": 密钥数据的字节串（可选）,
                  "extended": 扩展数据的字节串（可选）,
              }
        示例：
            def do_func(cracker: CrackerS1, count: int) -> dict[str, bytes]:
                plaintext = random.randbytes(aes_data_len)
                status, ciphertext = cracker.uart_transmit_receive(cmd_aes_enc + plaintext_data, rx_count= 12)
                return {
                    "plaintext": plaintext,
                    "ciphertext": ciphertext,
                    "key": aes_key,
                }
    :type do_func: typing.Callable[[CrackerS1, int], dict[str, bytes]]
    :param finish_func:
        结束后处理函数，该函数接受一个CrackerS1实例作为参数，用于在采集结束后进行清理操作。
        参数说明：
            - cracker: (CrackerS1) Cracker实例
            - 返回值：无
        示例：
            def finish_func(cracker: CrackerS1) -> None:
                cracker.nut_voltage_disable()
                cracker.nut_clock_disable()
                cracker.uart_io_disable()
    :type finish_func: typing.Callable[[CrackerS1], None]
    :param kwargs: 其他Acquisition的关键字参数
    :type kwargs: dict
    :return: Acquisition实例
    """

    class AnonymousAcquisition(Acquisition):
        def __init__(self, **_kwargs):
            _kwargs['cracker'] = cracker
            super().__init__(**_kwargs)

        def init(self):
            init_func(cracker)

        def do(self, count: int):
            return do_func(cracker, count)

        def finish(self):
            finish_func(cracker)

    return AnonymousAcquisition(**kwargs)


def simple_glitch_acq(
        cracker: CrackerG1,
        init_func: typing.Callable[[CrackerG1], None] = lambda cracker: None,
        do_func: typing.Callable[[CrackerG1, int], dict[str, bytes]] = lambda cracker, count: {},
        finish_func: typing.Callable[[CrackerG1], None] = lambda cracker: None, **kwargs) -> GlitchAcquisition:
    """
    创建一个简单的GlitchAcquisition子类实例。

    :param cracker: Cracker实例
    :type cracker: CrackerG1
    :param init_func:
        初始化函数，该函数接受一个CrackerG1实例作为参数，用于在采集开始前进行初始化操作。
        参数说明：
            - cracker: (CrackerG1) Cracker实例
            - 返回值：无
        示例：
            cracker.nut_voltage_enable()
            cracker.nut_voltage(3.3)
            cracker.nut_clock_enable()
            cracker.nut_clock_freq('8M')
            cracker.uart_io_enable()
            status, ret = cracker.uart_transmit_receive(cmd_set_aes_enc_key + aes_key, timeout=1000, rx_count=6)
    :type init_func: typing.Callable[[CrackerG1], None]
    :param do_func:
        执行函数，具体的采集逻辑实现，该函数接受一个CrackerG1实例和当前采集计数作为参数，返回一个包含采集数据的字典。
        参数说明：
            - cracker: (CrackerG1) Cracker实例
            - count: (int) 当前采集计数，从0开始
            - 返回值：dict[str, bytes]，包含采集数据的字典，格式如下：
              {
                  "plaintext": 明文数据的字节串,
                  "ciphertext": 密文数据的字节串,
                  "key": 密钥数据的字节串（可选）,
                  "extended": 扩展数据的字节串（可选）,
              }
        示例：
            def do_func(cracker: CrackerG1, count: int) -> dict[str, bytes]:
                plaintext = random.randbytes(aes_data_len)
                status, ciphertext = cracker.uart_transmit_receive(cmd_aes_enc + plaintext_data, rx_count= 12)
                return {
                    "plaintext": plaintext,
                    "ciphertext": ciphertext,
                    "key": aes_key,
                }
    :type do_func: typing.Callable[[CrackerG1, int], dict[str, bytes]]
    :param finish_func:
        结束后处理函数，该函数接受一个CrackerG1实例作为参数，用于在采集结束后进行清理操作。
        参数说明：
            - cracker: (CrackerG1) Cracker实例
            - 返回值：无
        示例：
            def finish_func(cracker: CrackerG1) -> None:
                cracker.nut_voltage_disable()
                cracker.nut_clock_disable()
                cracker.uart_io_disable()
    :type finish_func: typing.Callable[[CrackerG1], None]
    :param kwargs: 其他Acquisition的关键字参数
    :type kwargs: dict
    :return: Acquisition实例
    """
    class AnonymousAcquisition(GlitchAcquisition):
        def __init__(self, **_kwargs):
            _kwargs['cracker'] = cracker
            super().__init__(**_kwargs)

        def init(self):
            init_func(cracker)

        def do(self, count: int):
            return do_func(cracker, count)

        def finish(self):
            finish_func(cracker)

    return AnonymousAcquisition(**kwargs)


__all__ = ["simple_acq", "simple_glitch_acq"]



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib
import sqlite3

import traitlets
from anywidget import AnyWidget

from cracknuts import logger


class GlitchTestResultAnalysisPanel(AnyWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "GlitchTestResultAnalysisWidget.js"
    _css = ""

    selected_status = traitlets.List([0, 1, 2, 3]).tag(sync=True)
    glitch_test_result = traitlets.List([]).tag(sync=True)

    def __init__(self, glitch_test_result_path: str):
        super().__init__()
        self._logger = logger.get_logger(self)
        self.glitch_test_result_path = glitch_test_result_path
        self._conn = sqlite3.connect(self.glitch_test_result_path)
        self.cursor = self._conn.cursor()
        self.load_data()

    def load_data(self):
        query = f"SELECT * FROM glitch_result where status in ({", ".join(map(str, self.selected_status))})"
        self.cursor.execute(query)
        rows = self.cursor.fetchall()

        # print([dict(zip([column[0] for column in self.cursor.description], row)) for row in rows][0])
        self.glitch_test_result = [dict(zip([column[0] for column in self.cursor.description], row)) for row in rows]

    @traitlets.observe("selected_status")
    def on_selected_status_change(self, change):
        self.load_data()



# Copyright 2024 CrackNuts. All rights reserved.

import typing

from traitlets import traitlets

from cracknuts.jupyter.panel import MsgHandlerPanelWidget
from cracknuts.jupyter.ui_sync import observe_interceptor
from cracknuts.cracker.cracker_g1 import (CrackerG1, wave_4m, wave_8m, wave_10m, wave_16m, wave_20m, wave_40m, wave_80m,
                                          get_clock_from_wave_length)

class ConfigG1Glitch(MsgHandlerPanelWidget):

    glitch_vcc_normal = traitlets.Float(3.3).tag(sync=True)
    glitch_vcc_config_wait = traitlets.Int(1).tag(sync=True)
    glitch_vcc_config_level = traitlets.Float(3.3).tag(sync=True)
    glitch_vcc_config_count = traitlets.Int(1).tag(sync=True)
    glitch_vcc_config_delay = traitlets.Int(0).tag(sync=True)
    glitch_vcc_config_repeat = traitlets.Int(1).tag(sync=True)

    glitch_gnd_normal = traitlets.Float(0.0).tag(sync=True)
    glitch_gnd_config_wait = traitlets.Int(1).tag(sync=True)
    glitch_gnd_config_level = traitlets.Float(0.0).tag(sync=True)
    glitch_gnd_config_count = traitlets.Int(1).tag(sync=True)
    glitch_gnd_config_delay = traitlets.Int(0).tag(sync=True)
    glitch_gnd_config_repeat = traitlets.Int(1).tag(sync=True)

    _glitch_clock_embed_wave: dict[str, list[float]] = {
        "4M": wave_4m,
        "8M": wave_8m,
        "10M": wave_10m,
        "16M": wave_16m,
        "20M": wave_20m,
        "40M": wave_40m,
        "80M": wave_80m
    }
    # These two attributes are used to synchronize waveform related settings made in Python code with the Jupyter interface.
    _glitch_clock_wave_normal: list[float] = wave_8m
    _glitch_clock_config_wave_glitch: list[float] = wave_4m
    glitch_clock_normal_freq_items: list[dict[str, typing.Any]] = traitlets.List([{
        "name":  f"{name}-[{', '.join(map(str, wave))}]",
        "tooltip": f"[{', '.join(map(str, wave))}]",
        "wave": wave
    } for name, wave in _glitch_clock_embed_wave.items()]).tag(sync=True)
    glitch_clock_normal_selected_freq_item_idx: int = traitlets.Int(1).tag(sync=True) # default 8m
    glitch_clock_glitch_freq_items: list[dict[str, typing.Any]] = traitlets.List([{
        "name": f"{name}-[{', '.join(map(str, wave))}]",
        "tooltip": f"[{', '.join(map(str, wave))}]",
        "wave": wave
    } for name, wave in _glitch_clock_embed_wave.items()]).tag(sync=True)
    glitch_clock_glitch_selected_freq_item_idx: int = traitlets.Int(1).tag(sync=True) # default 8m

    @property
    def glitch_clock_wave_normal(self) -> list[float]:
        return self._glitch_clock_wave_normal

    @glitch_clock_wave_normal.setter
    def glitch_clock_wave_normal(self, value: list[float]) -> None:
        self._glitch_clock_wave_normal = value
        for idx, item in enumerate(self.glitch_clock_normal_freq_items):
            if value == item["wave"]:
                self.glitch_clock_normal_selected_freq_item_idx = idx
                break

    @property
    def glitch_clock_config_wave_glitch(self) -> list[float]:
        return self._glitch_clock_config_wave_glitch

    @glitch_clock_config_wave_glitch.setter
    def glitch_clock_config_wave_glitch(self, value: list[float]) -> None:
        for idx, item in enumerate(self.glitch_clock_glitch_freq_items):
            if value == item["wave"]:
                self.glitch_clock_glitch_selected_freq_item_idx = idx
                self._glitch_clock_config_wave_glitch = value
                return
        self._glitch_clock_freq_add(value)

    glitch_clock_config_wait: int = traitlets.Int(1).tag(sync=True)
    glitch_clock_config_delay: int = traitlets.Int(0).tag(sync=True)
    glitch_clock_config_repeat: int = traitlets.Int(1).tag(sync=True)


    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.cracker: CrackerG1 = kwargs["cracker"]
        self.reg_msg_handler("glitchVCCForceButton", "onClick", self.glitch_vcc_force)
        self.reg_msg_handler("glitchGNDForceButton", "onClick", self.glitch_gnd_force)
        self.reg_msg_handler("glitchClockFreqAddButton", "onClick", self.glitch_clock_freq_add)
        self.reg_msg_handler("glitchClockForceButton", "onClick", self.glitch_clock_force)

    def glitch_vcc_force(self, args: dict[str, typing.Any]):
        self.cracker.glitch_vcc_force()

    def glitch_gnd_force(self, args: dict[str, typing.Any]):
        self.cracker.glitch_gnd_force()

    def glitch_clock_freq_add(self, args: dict[str, typing.Any]):
        if "wave" in args.keys():
            new_wave = args["wave"]
            self._glitch_clock_freq_add(new_wave)

    def glitch_clock_force(self, args: dict[str, typing.Any]):
        self.cracker.glitch_clock_force()

    def _glitch_clock_freq_add(self, new_wave):
        glitch_clock_freq_items = [item for item in self.glitch_clock_glitch_freq_items]
        glitch_clock_freq_items.append({
            "name": f"C-{len(glitch_clock_freq_items) - len(self._glitch_clock_embed_wave)}-[{", ".join(map(str, new_wave))}]",
            "tooltip": f"[{", ".join(map(str, new_wave))}]",
            "wave": new_wave
        })
        self.glitch_clock_glitch_freq_items = glitch_clock_freq_items
        self.glitch_clock_glitch_selected_freq_item_idx = len(glitch_clock_freq_items) - 1
        self._glitch_clock_config_wave_glitch = new_wave

    def glitch_clock_freq_selected(self, args: dict[str, typing.Any]):
        if "selected" in args.keys():
            selected_clock_freq_item_idx = args["selected"]
            selected_clock_freq_item = self.glitch_clock_glitch_freq_items[selected_clock_freq_item_idx]
            if selected_clock_freq_item:
                self.glitch_clock_config_wave_glitch = selected_clock_freq_item["wave"]
                self.cracker.glitch_clock_config(
                    wave=self.glitch_clock_config_wave_glitch,
                    wait=self.glitch_clock_config_wait,
                    delay=self.glitch_clock_config_delay,
                    repeat=self.glitch_clock_config_repeat
                )


    def get_glitch_clock_normal_freq(self):
        return get_clock_from_wave_length(len(self.glitch_clock_wave_normal))

    def get_glitch_clock_glitch_freq(self):
        return get_clock_from_wave_length(len(self.glitch_clock_config_wave_glitch))

    @traitlets.observe("glitch_vcc_normal")
    @observe_interceptor
    def glitch_vcc_normal_changed(self, change):
        self.cracker.glitch_vcc_normal(change.get("new"))

    @traitlets.observe("glitch_vcc_config_level")
    @observe_interceptor
    def glitch_vcc_config_level_changed(self, change):
        self.cracker.glitch_vcc_config(
            self.glitch_vcc_config_wait,
            change.get("new"),
            self.glitch_vcc_config_count,
            self.glitch_vcc_config_delay,
            self.glitch_vcc_config_repeat,
        )

    @traitlets.observe("glitch_vcc_config_wait")
    @observe_interceptor
    def glitch_vcc_config_wait_changed(self, change):
        self.cracker.glitch_vcc_config(
            change.get("new"),
            self.glitch_vcc_config_level,
            self.glitch_vcc_config_count,
            self.glitch_vcc_config_delay,
            self.glitch_vcc_config_repeat,
        )

    @traitlets.observe("glitch_vcc_config_count")
    @observe_interceptor
    def glitch_vcc_config_count_changed(self, change):
        self.cracker.glitch_vcc_config(
            self.glitch_vcc_config_wait,
            self.glitch_vcc_config_level,
            change.get("new"),
            self.glitch_vcc_config_delay,
            self.glitch_vcc_config_repeat,
        )

    @traitlets.observe("glitch_vcc_config_repeat")
    @observe_interceptor
    def glitch_vcc_config_repeat_changed(self, change):
        self.cracker.glitch_vcc_config(
            self.glitch_vcc_config_wait,
            self.glitch_vcc_config_level,
            self.glitch_vcc_config_count,
            self.glitch_vcc_config_delay,
            change.get("new"),
        )

    @traitlets.observe("glitch_vcc_config_delay")
    @observe_interceptor
    def glitch_vcc_config_delay_changed(self, change):
        self.cracker.glitch_vcc_config(
            self.glitch_vcc_config_wait,
            self.glitch_vcc_config_level,
            self.glitch_vcc_config_count,
            change.get("new"),
            self.glitch_vcc_config_repeat,
        )

    @traitlets.observe("glitch_gnd_normal")
    @observe_interceptor
    def glitch_gnd_normal_changed(self, change):
        self.cracker.glitch_gnd_normal(change.get("new"))

    @traitlets.observe("glitch_gnd_config_wait")
    @observe_interceptor
    def glitch_gnd_config_wait_changed(self, change):
        self.cracker.glitch_gnd_config(
            change.get("new"),
            self.glitch_gnd_config_level,
            self.glitch_gnd_config_count,
            self.glitch_gnd_config_delay,
            self.glitch_gnd_config_repeat,
        )

    @traitlets.observe("glitch_gnd_config_level")
    @observe_interceptor
    def glitch_gnd_config_level_changed(self, change):
        self.cracker.glitch_gnd_config(
            self.glitch_gnd_config_wait,
            change.get("new"),
            self.glitch_gnd_config_count,
            self.glitch_gnd_config_delay,
            self.glitch_gnd_config_repeat,
        )

    @traitlets.observe("glitch_gnd_config_count")
    @observe_interceptor
    def glitch_gnd_config_count_changed(self, change):
        self.cracker.glitch_gnd_config(
            self.glitch_gnd_config_wait,
            self.glitch_gnd_config_level,
            change.get("new"),
            self.glitch_gnd_config_delay,
            self.glitch_gnd_config_repeat,
        )

    @traitlets.observe("glitch_gnd_config_delay")
    @observe_interceptor
    def glitch_gnd_config_changed(self, change):
        self.cracker.glitch_gnd_config(
            self.glitch_gnd_config_wait,
            self.glitch_gnd_config_level,
            self.glitch_gnd_config_count,
            change.get("new"),
            self.glitch_gnd_config_repeat,
        )

    @traitlets.observe("glitch_gnd_config_repeat")
    @observe_interceptor
    def glitch_gnd_config_repeat_changed(self, change):
        self.cracker.glitch_gnd_config(
            self.glitch_gnd_config_wait,
            self.glitch_gnd_config_level,
            self.glitch_gnd_config_count,
            self.glitch_gnd_config_delay,
            change.get("new"),
        )

    @traitlets.observe("glitch_clock_config_wait")
    @observe_interceptor
    def glitch_clock_config_wait_changed(self, change):
        self.cracker.glitch_clock_config(
            wave=self._glitch_clock_config_wave_glitch,
            wait=change.get("new"),
            delay=self.glitch_clock_config_delay,
            repeat=self.glitch_clock_config_repeat,
        )

    @traitlets.observe("glitch_clock_config_delay")
    @observe_interceptor
    def glitch_clock_config_delay_changed(self, change):
        self.cracker.glitch_clock_config(
            wave=self._glitch_clock_config_wave_glitch,
            wait=self.glitch_clock_config_wait,
            delay=change.get("new"),
            repeat=self.glitch_clock_config_repeat,
        )

    @traitlets.observe("glitch_clock_config_repeat")
    @observe_interceptor
    def glitch_clock_config_repeat_changed(self, change):
        self.cracker.glitch_clock_config(
            wave=self._glitch_clock_config_wave_glitch,
            wait=self.glitch_clock_config_wait,
            delay=self.glitch_clock_config_delay,
            repeat=change.get("new"),
        )

    @traitlets.observe("glitch_clock_glitch_selected_freq_item_idx")
    @observe_interceptor
    def glitch_clock_glitch_selected_freq_item_idx_changed(self, change):
        self._glitch_clock_config_wave_glitch = self.glitch_clock_glitch_freq_items[change.get("new")]["wave"]
        self.cracker.glitch_clock_config(
            wave=self._glitch_clock_config_wave_glitch,
            wait=self.glitch_clock_config_wait,
            delay=self.glitch_clock_config_delay,
            repeat=self.glitch_clock_config_repeat,
        )